import 'package:brick_offline_first/src/offline_queue/graphql/graphql_request_sqlite_cache_manager.dart';
import 'package:brick_offline_first/src/offline_queue/graphql/graphql_offline_queue_link.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:gql_exec/gql_exec.dart';
import 'package:mockito/mockito.dart';
import 'package:sqflite_common/sqlite_api.dart';
import 'package:sqflite_common_ffi/sqflite_ffi.dart';
import 'package:gql/language.dart';
import '../__helpers__.dart';

void main() {
  TestWidgetsFlutterBinding.ensureInitialized();
  sqfliteFfiInit();

  const query = '''query UpsertPerson {
        upsertPerson {
          firstName
        }
      }''';

  const response = Response(
    data: <String, dynamic>{'firstName': 'Beavis'},
  );

  final request = Request(
    operation: Operation(document: parseString(query), operationName: 'UpsertPerson'),
    variables: const <String, dynamic>{'firstName': 'Beavis'},
  );

  group('GraphqlOfflineQueueLink', () {
    final requestManager = GraphqlRequestSqliteCacheManager(
      inMemoryDatabasePath,
      databaseFactory: databaseFactoryFfi,
    );

    setUpAll(() async {
      await requestManager.migrate();
    });

    tearDown(() async {
      final requests = await requestManager.unprocessedRequests();
      final requestsToDelete = requests.map((request) {
        return requestManager.deleteUnprocessedRequest(request[GRAPHQL_JOBS_PRIMARY_KEY_COLUMN]);
      });

      await Future.wait(requestsToDelete);
    });

    test('verify link request made', () async {
      final mockLink = stubGraphqlLink({});
      final client = GraphqlOfflineQueueLink(mockLink, requestManager);

      await client.request(request).first;

      verify(
        mockLink.request(request),
      ).called(1);
    });

    test('#send forwards to inner client', () async {
      final mockLink = MockLink();

      final client = GraphqlOfflineQueueLink(mockLink, requestManager);

      when(
        mockLink.request(request),
      ).thenAnswer(
        (_) => Stream.fromIterable([response]),
      );

      expect(await client.request(request).first, response);
    });

    test('Query / Subscriptions are not tracked', () async {
      final mockLink = stubGraphqlLink({});
      final client = GraphqlOfflineQueueLink(mockLink, requestManager);

      await client
          .request(
            Request(
              operation: Operation(
                document: parseString('''query {
                  helloWorld {
                    name
                  }
                }'''),
              ),
            ),
          )
          .first;
      expect(await requestManager.unprocessedRequests(), isEmpty);
    });

    test('request is stored in SQLite', () async {
      final mockLink = stubGraphqlLink({}, errors: ['Unavailable']);
      final client = GraphqlOfflineQueueLink(mockLink, requestManager);

      await client
          .request(
            Request(
              operation: Operation(
                document: parseString('''mutation {}'''),
              ),
            ),
          )
          .first;
      expect(await requestManager.unprocessedRequests(), hasLength(1));
    });

    test('.isMutation', () async {
      final mutationRequest = Request(
        operation: Operation(
          document: parseString('''mutation {}'''),
        ),
      );

      final queryRequest = Request(
        operation: Operation(
          document: parseString('''query {}'''),
        ),
      );

      final subscriptionRequest = Request(
        operation: Operation(
          document: parseString('''subscription {}'''),
        ),
      );
      expect(GraphqlOfflineQueueLink.isMutation(mutationRequest), isTrue);
      expect(GraphqlOfflineQueueLink.isMutation(queryRequest), isFalse);
      expect(GraphqlOfflineQueueLink.isMutation(subscriptionRequest), isFalse);
    });

    test('request deletes after a successful response', () async {
      final mockLink = MockLink();

      final client = GraphqlOfflineQueueLink(mockLink, requestManager);

      when(
        mockLink.request(request),
      ).thenAnswer(
        (_) => Stream.fromIterable([response]),
      );

      client.request(
        Request(
          operation: Operation(
            document: parseString('''mutation {}'''),
          ),
        ),
      );

      expect(await requestManager.unprocessedRequests(), isEmpty);
    });

    test('request increments after a unsuccessful response', () async {
      final mockLink = stubGraphqlLink({}, errors: ['Unsuccessful']);

      final client = GraphqlOfflineQueueLink(mockLink, requestManager);
      final mutationRequest = Request(
        operation: Operation(
          document: parseString('''mutation {}'''),
          operationName: 'fakeMutate',
        ),
      );
      await client.request(mutationRequest).first;

      client.request(mutationRequest);

      var unprocessedRequests = await requestManager.unprocessedRequests();

      expect(unprocessedRequests.first[GRAPHQL_JOBS_ATTEMPTS_COLUMN], 1);

      await client.request(mutationRequest).first;

      unprocessedRequests = await requestManager.unprocessedRequests();
      expect(unprocessedRequests.first[GRAPHQL_JOBS_ATTEMPTS_COLUMN], 2);
    });

    test('request creates and does not delete after an unsuccessful response', () async {
      final mockLink = stubGraphqlLink({}, errors: ['Unknown error']);
      final client = GraphqlOfflineQueueLink(mockLink, requestManager);
      final mutationRequest = Request(
        operation: Operation(
          document: parseString('''mutation {}'''),
          operationName: 'fakeMutate',
        ),
      );

      await client.request(mutationRequest).first;

      client.request(mutationRequest);

      expect(await requestManager.unprocessedRequests(), hasLength(1));
    });

    test('request is not deleted after sending to a misconfigured client', () async {
      final link = stubGraphqlLink({}, errors: ['Misconfigured']);

      final client = GraphqlOfflineQueueLink(link, requestManager);
      const document = '''mutation {
            hello{
              hi
            }
          }''';
      final mutationRequest1 = Request(
        operation: Operation(
          document: parseString(document),
          operationName: 'fakeMutate',
        ),
      );

      final mutationRequest2 = Request(
        operation: Operation(
          document: parseString(document),
          operationName: 'fakeMutate',
        ),
        variables: const <String, dynamic>{'j': 16},
      );

      final mutationRequest3 = Request(
        operation: Operation(
          document: parseString(document),
          operationName: 'fakeMutate',
        ),
        variables: const <String, dynamic>{'k': 14},
      );

      await client.request(mutationRequest1).first;

      expect(await requestManager.unprocessedRequests(), hasLength(1));

      await client.request(mutationRequest2).first;

      expect(await requestManager.unprocessedRequests(), hasLength(2));

      await client.request(mutationRequest3).first;

      expect(await requestManager.unprocessedRequests(), hasLength(3));
    });
  });
}
